/*
    cpuspeed2.cc 2004.11.12 09:55

    Copyright 2004

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    cpuspeed2.cc

    Author:
        Bill Waddington <william.waddington@beezmo.com>

    Based on cpuspeed.cc

        Author:
            Carl Thompson (carlthompson.net)

        Contributors:
            Enrico Tassi

    This program is only for computers with Linux kernels compiled with
    CPUFreq.  You must have a CPU that supports frequency and/or voltage
    scaling via CPUFreq to use this program.  Your kernel must be compiled to
    support the "userspace" CPUFreq governor and the "sysfs" interface.

    In this program, cpuspeed's support for the 2.4 kernel series has been
    dropped.  Other changes: automated search for /proc/acpi files, separate
    policies for AC, battery, and low battery operation, modified command
    line options, and enhanced APM support.

    Tested on my IBM Thinkpad R40 Centrino laptop running Fedora Core 2 with
    a 2.6.8.1 kernel.
*/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/errno.h>
#include <stdarg.h>
#include <time.h>
#include <sys/utsname.h>
#include <dirent.h>
#include <ctype.h>

#ifndef NAME
#  define NAME "cpuspeed2"
#endif

#define VERSION "0.52"

// display an error message and exit the program
void
die(bool system_error, const char *fmt, ...)
{
    fprintf(stderr, "Error: ");

    va_list ap;
    // get variable argument list passed
    va_start(ap, fmt);
    // display message passed on stderr
    vfprintf(stderr, fmt, ap);
    va_end(ap);

    fprintf(stderr, "\n");
    if (system_error)
        fprintf(stderr, "Error: %s\n", strerror(errno));

    exit(1);
}

// Maximum number of speed steps.  Must be 2 or greater
#define MAX_SPEEDS  20

// To activate debug printings
#ifdef DEBUG
#   define errprintf(A...) fprintf(stderr,A)
#else
#   define errprintf(A...)
#endif

//APM and ACPI file paths and keys
#define SYSFS_AVAILABLE_SPEEDS_FILE \
    "/sys/devices/system/cpu/cpu0/cpufreq/scaling_available_frequencies"

#define SYSFS_CURRENT_SPEED_FILE \
    "/sys/devices/system/cpu/cpu0/cpufreq/scaling_setspeed"

#define SYSFS_GOVERNOR_FILE \
    "/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor"

#define SYSFS_USERSPACE \
    "userspace"

#define PROC_APM_ACDC_FILE \
    "/proc/apm"

#define PROC_ACPI_DIR \
    "/proc/acpi"

#define PROC_ACPI_AC_STATE_BASE \
    "/proc/acpi/ac_adapter/"
#define PROC_ACPI_AC_STATE_FILE \
    "state"

#define PROC_ACPI_BATT_STATE_BASE \
    "/proc/acpi/battery/"
#define PROC_ACPI_BATT_STATE_FILE \
    "state"

#define PROC_ACPI_BATT_INFO_BASE \
    "/proc/acpi/battery/"
#define PROC_ACPI_BATT_INFO_FILE \
    "info"

#define PROC_ACPI_TEMPERATURE_BASE \
    "/proc/acpi/thermal_zone/"
#define PROC_ACPI_TEMPERATURE_FILE \
    "temperature"

#define ACPI_BATT_LAST_MAX_CAP_KEY \
    "last full capacity"

#define ACPI_BATT_REM_CAP_KEY \
    "remaining capacity"

#define ACPI_TEMPERATURE_KEY \
    "temperature"

// will be filled in by searching appropriate path
// may be overridden by cmd line options
char * proc_acpi_ac_state_file = 0;
char * proc_acpi_batt_state_file = 0;
char * proc_acpi_batt_info_file = 0;
char * proc_acpi_temperature_file = 0;

// place for constructed acpi paths (if needed)
#define MAX_PATH_LEN 63 
char acpi_ac_state_path[MAX_PATH_LEN + 1];
char acpi_batt_state_path[MAX_PATH_LEN + 1];
char acpi_batt_info_path[MAX_PATH_LEN + 1];
char acpi_temperature_path[MAX_PATH_LEN + 1];

// build acpi temperature, ac, and battery file paths
// some code borrowed from Gnome Battstat appled acpi-linux.c
void find_path(char *base, char *file, char *fullpath, int maxlen)
{
    DIR *basedir;
    struct dirent *dir_entry;

    basedir = opendir(base);
    if(!basedir)
        die(false, "Unable to open %s directory\n", base);
    while((dir_entry = readdir(basedir))) {
        if(dir_entry->d_name[0] != '.') {
            strncpy(fullpath, base, maxlen - (strlen(base) + 1));
            strncat(fullpath, dir_entry->d_name, \
                maxlen - (strlen(fullpath) + strlen(dir_entry->d_name) + 1));
            strncat(fullpath, "/", maxlen - (strlen(fullpath) + 1 + 1));
            strncat(fullpath, file, maxlen - (strlen(fullpath) + strlen(file) + 1));
        }
    }
    closedir(basedir);
}

// defines what info we care about for each speed step
struct
{
    unsigned khz;
    //unsigned volts;
    //unsigned fsb;
} speeds[MAX_SPEEDS + 1];

// read a line from a file
void
read_line(const char * filename, char * line, unsigned len)
{
    FILE * fp = fopen(filename, "r");
    if (!fp)
        die(true, "Could not open file for reading: %s", filename);
    if ( (!fgets(line, len, fp)) )
        die(true, "Could not read from file: %s", filename);
    fclose(fp);
}

// write a line to a file
void
write_line(const char * filename, const char *fmt, ...)
{
    FILE * fp = fopen(filename, "w+");
    if (!fp)
        die(true, "Could not open file for writing: %s", filename);
    va_list ap;
    // get variable argument list passed
    va_start(ap, fmt);
    if (vfprintf(fp, fmt, ap) < 0)
        die(true, "Could not write to file: %s", filename);
    va_end(ap);
    fclose(fp);
}

// read an integer value from a file
unsigned
read_value(const char * filename)
{
    char line[256];
    read_line(filename, line, sizeof line);
    return atoi(line);
}

// get int value - any base
long
get_int(const char * s)
{
    char * end;
    long r = strtol(s, &end, 0);
    if (*end != '\0')
        die(false, "Not an integer: %s", s);
    if (errno == ERANGE)
        die(false, "Number is out of range: %s", s);

    return r;
}

// convert a string to lower case
void string_tolower(char *string)
{
    while(*string != '\0') {
        *string = (char)tolower((int)*string);
        string++;
    }
}

// get the current CPU speed
unsigned get_speed()
{
    return read_value(SYSFS_CURRENT_SPEED_FILE);
}

// scan file for "key" and return integer value
// file data case insensitive - key must be lower case
int get_int_key_value(const char *filename, const char *key)
{
    char line[256], *p;
    FILE * fp = fopen(filename, "r");
    if (!fp)
        die(true, "Could not open file for reading: %s", filename);
    while (fgets(line, sizeof(line), fp)) {
        string_tolower(line); // just in "case"
        if(strstr(line, key)) {
            fclose(fp);
            p = strpbrk(line, "0123456789");
            if (!p)
                die(false, "No value associated with key \"%s\" in file %s\n",
                    key, filename);
            return strtol(p, NULL, 10);
        }
    }
    fclose(fp);
    die(false, "Could not find key \"%s\" in file %s\n", key, filename);
    return 0 ; // won't get here 
}


// get the current CPU speed index
// speeds[] must be valid prior to calling!
unsigned get_speed_index()
{
    unsigned speed_index = 0;
    unsigned current_speed = get_speed();
    do {
        if(speeds[speed_index].khz == current_speed) 
            return(speed_index);
    } while(speeds[speed_index++].khz != 0);
                                                                                                    
    die(false, "Could not determine speed index!");
    // won't get here
    return 0;
}

// set the current CPU speed
void set_speed(unsigned value)
{
    errprintf("Setting speed to: %uKHz\n", value);
    write_line(SYSFS_CURRENT_SPEED_FILE, "%u\n", value);
    // give CPU / chipset voltage time to settle down
    usleep(10000);
}

// This code smoothly transitions the CPU speed from 'current' to 'target'
// instead of jumping directly to the new speed because some AMD Mobile
// Athlon parts seem to choke on large differentials causing kernel panics.
void set_speed(unsigned current, unsigned target)
{
    if (current == target)
        return;
    int delta = (current > target) ? -1 : 1;
    do
    {
        current += delta;
        set_speed(speeds[current].khz);
    }
    while (current != target);
}

// lowest speed step
unsigned last_step;

// get the speed steps supported by the CPU
// read directly from /sys/devices/...
// the rest of the code wants highest first 
void
get_supported_speeds()
{
    int temp_speeds[MAX_SPEEDS];
    int i, num_speeds, spd;
    char line[256];
    char *curpos=line;

    read_line(SYSFS_AVAILABLE_SPEEDS_FILE, line, sizeof line);
    num_speeds=0;
    do {
        spd = strtol(curpos, &curpos, 0);
        temp_speeds[num_speeds++] = spd;
    } while (spd && (num_speeds < MAX_SPEEDS));
    num_speeds--;

    // swizzle speed list if necesary (it always is...)
    if(temp_speeds[0] < temp_speeds[num_speeds-1]) {
        for(i=0; i<num_speeds; i++)
            speeds[i].khz = temp_speeds[(num_speeds - 1) - i];
    } else {
        for(i=0; i<num_speeds; i++)
            speeds[i].khz = temp_speeds[i];
    }

    speeds[num_speeds].khz=0;

    // the last step is the lowest found speed
    last_step = num_speeds - 1;

#ifdef DEBUG
    errprintf("Available speeds:\n");
    for (i = 0; speeds[i].khz; i++)
        errprintf(" %2u: %9uKHz\n", i, speeds[i].khz);
#endif

#ifdef DEBUG
        usleep(150000);
#endif

    // go back to maximum speed because program expects to start there
    set_speed(get_speed_index(), 0);
}

// gets the elapsed total time and elapsed idle time since it was last called
void
get_times(unsigned long & total_elapsed, unsigned long & idle_elapsed)
{
    char what[32];
    unsigned long user_time, nice_time, system_time, idle_time, total_time;
    static unsigned long last_total_time = 0, last_idle_time = 0;

    bool found;
    char line[256];
    FILE * fp = fopen("/proc/stat", "r");
    if (!fp)
        die(true, "Could not open /proc/stat for reading!");
    while ( (found = fgets(line, sizeof line, fp)) )
    {
        sscanf(line, "%s %lu %lu %lu %lu", what, &user_time, &nice_time,
               &system_time, &idle_time);
        if (!strcmp(what, "cpu0"))
            break;
    }
    fclose(fp);

    if (!found)
        die(false, "Could not find entry for cpu0 in /proc/stat!?");

    // count nice time as idle time
    idle_time += nice_time;

    total_time = user_time + system_time + idle_time;
    total_elapsed = total_time - last_total_time;
    last_total_time = total_time;
    idle_elapsed = idle_time - last_idle_time;
    last_idle_time = idle_time;

    //errprintf("time: %lu    idle: %lu\n", total_elapsed, idle_elapsed);
}

// resets the elapsed total time and elapsed idle time counters
void reset_times()
{
    unsigned long dummy1, dummy2;
    get_times(dummy1, dummy2);
}

// are we currently dynamically scaling the CPU or at min or max?
enum Mode { MIN, MAX, DYNAMIC } mode = DYNAMIC;

// we have separate policies for ac, battery, and low battery
// can be set via .config or overridden w/signals
Mode ac_mode = DYNAMIC;
Mode batt_mode = MIN;
Mode low_batt_mode = MIN;

// list of valid policies and mode values
struct
{
    char * string;
    Mode mode;
} modelist[] =
{
    "minimum", MIN,
    "maximum", MAX,
    "dynamic", DYNAMIC,
    NULL, (Mode)0,
};
  
// set policy
bool set_policy( char *policy, Mode *mode)
{
    int i = 0;
    while(modelist[i].string) {
        if(!strcmp(policy, modelist[i].string)) {
            *mode = modelist[i].mode;
            return true;
        }
        i++;
    }
    return false;
}

// are external signals controlling our policy or are we free to
// use defaults or .config settings?
enum SigMode { SIG_DYNAMIC, SIG_FREE, SIG_MIN, SIG_MAX } sig_mode = SIG_DYNAMIC;

// if battery % remaining below this, we are in low batt mode
unsigned low_batt_threshold  = 20;

// if CPU idle percentage is below this, CPU will be set to fastest speed
unsigned clock_up_idle_fast = 10;

// if CPU idle percentage is below this, CPU will be set to next higher speed
// if CPU idle percentage is above this, CPU will be set to next lower speed
unsigned idle_threshold = 25;

// throttle down CPU to minimum if temperature is above this value (C)
// disable on command line by setting this temp to 0
// default to 75 - I HAVE NO IDEA IF THIS IS AN APPROPRIATE TEMP!
unsigned max_temperature = 75;

// apm or acpi mode?
bool using_acpi = false;

// return !0 if AC is disconnected
// use apm or acpi appropriate interrogation
bool ac_offline()
{
    char line[256];
    unsigned ac_online, ac_offline = false;

    if(using_acpi) {
        if(*proc_acpi_ac_state_file == '\0')
            return false;
        read_line(proc_acpi_ac_state_file, line, sizeof line);
        string_tolower(line);
        if (strstr(line, "off-line") || strstr(line, "on-line")) {
            if(strstr(line, "off-line"))
                ac_offline = true;
        } else
           die(false, "invalid ac line status in %s\n", proc_acpi_ac_state_file);
    } else {
        read_line(PROC_APM_ACDC_FILE, line, sizeof line);
        sscanf(line, "%*s %*s %*s %i", &ac_online);
        if(!ac_online)
            ac_offline = true;
    }
#ifdef DEBUG
    if(ac_offline)
        errprintf("AC offline\n");
    else
        errprintf("AC online\n");
#endif
    return ac_offline; 
}

// get battery charge percentage
// apm or acpi as appropriate
unsigned current_batt_capacity()
{
    char line[256];
    unsigned pc_remaining, last_max_cap, mwh_remaining;

    if(using_acpi) {
        if((*proc_acpi_batt_info_file == '\0') ||
            (*proc_acpi_batt_state_file == '\0'))
            return 100;
        last_max_cap = get_int_key_value(proc_acpi_batt_info_file, \
            ACPI_BATT_LAST_MAX_CAP_KEY);
        mwh_remaining = get_int_key_value(proc_acpi_batt_state_file, \
            ACPI_BATT_REM_CAP_KEY);
        pc_remaining = (mwh_remaining * 100) / last_max_cap;
    } else {
        read_line(PROC_APM_ACDC_FILE, line, sizeof line);
        sscanf(line, "%*s %*s %*s %*s %*s %*s %i", &pc_remaining);
    }
    errprintf("Battery status: %d%% remaining - low battery threshold %d%%.\n",
        pc_remaining, low_batt_threshold);
    return pc_remaining;
}

// check for overtemp (if possible)
// need acpi mode.  max_temperature == 0 disables
bool over_temp()
{
    unsigned temp;

    if (max_temperature && using_acpi)
    {
        if(*proc_acpi_temperature_file == '\0')
            return false;
        temp = get_int_key_value(proc_acpi_temperature_file, ACPI_TEMPERATURE_KEY);
        errprintf("At the sound of the tone, the exact temperature is: %dC\n",
            temp);
        errprintf("Overtemp threshold is %dC\n", max_temperature);
        if (temp > max_temperature)
            return true;
    }
    return false;
}

// current ac/battery/temp status
enum { AC, BATT, LOW_BATT, OVERTEMP } power_therm_status = AC;

// we only bother to query battery/ac and temp status every
// LONG_INTERVAL_MULT passes
unsigned alarm_passes = 0;

// time between "ticks" in dynamic mode
unsigned interval = 20; // 2 seconds

// stretch interval x this in fixed freq modes
// only check ac, batt, thermal, every "long" interval in
// dynamic mode
#define LONG_INTERVAL_MULT 5

// this is _really_ paranoid - and trusts that
// this "integer" type will hold our max interval 
volatile sig_atomic_t working_interval;

// handles the periodic check of idle and setting CPU speed
void
alarm_handler(int)
{
    static unsigned new_speed = 0;
    static unsigned current_speed = 0;

    // get ac/batt status - every 10 passes or every pass
    // if mode != DYNAMIC (since we check less frequently)
    // over temp takes precedence over ac/dc status
    if(!(alarm_passes++ % LONG_INTERVAL_MULT) || (mode != DYNAMIC)) {
        if(over_temp())
            power_therm_status = OVERTEMP;
        else if(ac_offline())
            power_therm_status = (current_batt_capacity() < low_batt_threshold)
                ? LOW_BATT : BATT;
        else
            power_therm_status = AC;
    }

    // did a signal force our mode?
    // overtemp slowdown not available unless in
    // SIG_FREE mode
    switch(sig_mode)
    {
        case SIG_MIN:
            errprintf("Sig mode is MIN\n");
            mode = MIN;
            break;

        case SIG_MAX:
            errprintf("Sig mode is MAX\n");
            mode = MAX;
            break;

        case SIG_DYNAMIC:
            errprintf("Sig mode is DYNAMIC\n");
            mode = DYNAMIC;
            break;

        case SIG_FREE:
            errprintf("Sig mode is FREE\n");
            // set mode based on policies
            switch(power_therm_status)
            {
                case AC:
                    mode = ac_mode;
                    break;
                case BATT:
                    mode = batt_mode;
                    break;
                case LOW_BATT:
                   mode = low_batt_mode;
                   break;
                case OVERTEMP:
                   // always go to min to cool off
                   mode = MIN;
                   break;
            }
            break;
    }

    // figure out what our current speed should be
    switch(mode)
    {
        case DYNAMIC:
        {
            unsigned long elapsed_time, idle_time;
            unsigned idle_percent;

            errprintf("mode is DYANMIC\n");
            // get the elapsed and idle times since we last checked
            get_times(elapsed_time, idle_time);

            if (elapsed_time > 0)
            {
                idle_percent = idle_time * 100 / elapsed_time;

                if (idle_percent <= clock_up_idle_fast)
                    new_speed = 0;
                else if (idle_percent < idle_threshold && new_speed != 0)
                    new_speed--;
                else if (idle_percent > idle_threshold && speeds[new_speed + 1].khz != 0)
                    new_speed++;

                errprintf("Idle percent: %.2u   %4dKHz\n", idle_percent, speeds[new_speed].khz);
            }
            working_interval = interval;
            break;
        }
        // test less frequently when in fixed freq mode
        case MIN:
            errprintf("mode is MIN\n");
            new_speed = last_step;
            working_interval = interval * LONG_INTERVAL_MULT;
            reset_times();
            break;

        case MAX:
            errprintf("mode is MAX\n");
            new_speed = 0;
            working_interval = interval * LONG_INTERVAL_MULT;
            reset_times();
            break;
    }

    // if the last set speed is not what it currently should be, set it
    if (current_speed != new_speed)
    {
        errprintf("current speed: %uKHz     new speed: %uKHz\n",
                  speeds[current_speed].khz, speeds[new_speed].khz);
        set_speed(current_speed, new_speed);
        current_speed = new_speed;
    }
}

// handles the USR1 signal (stay at maximum performance)
void
usr1_handler(int)
{
    sig_mode = SIG_MAX;
}

// handles the USR2 signal (stay at minimum performance)
void
usr2_handler(int)
{
    sig_mode = SIG_MIN;
}

// handles the HUP signal (use policies)
void
hup_handler(int)
{
    sig_mode = SIG_DYNAMIC;
}

// handles the IO signal (use policies)
void
io_handler(int)
{
    sig_mode = SIG_FREE;
}

int
main(unsigned argc, char * argv[])
{
    bool daemonize = false;

    // parse argv
    for(unsigned i = 1; i < argc; i++)
    {
        if(!strcmp(argv[i], "-d"))
            daemonize = true;
        else if(!strcmp(argv[i], "-i"))
        {
            if (argc <= i + 1)
                die(false, "The -i option must be followed by an interval in tenths of a second\n");
            interval = get_int(argv[++i]);
            errprintf("(cmd line option) Interval is %u\n", interval);
        }
        else if(!strcmp(argv[i], "-p"))
        {
            if (argc <= i + 3)
                die(false, "The -p option must be followed by 2 integers\n");

            clock_up_idle_fast = get_int(argv[++i]);
            idle_threshold = get_int(argv[++i]);
            errprintf("(cmd line option) Triggers are %u %u\n",clock_up_idle_fast, idle_threshold);
        }
        else if (!strcmp(argv[i], "-t"))
        {
            if (argc <= i + 1)
                die(false, "The -t option must be followed by a temperature in degrees C\n");
            max_temperature = get_int(argv[++i]);
            errprintf("(cmd line option) High temperature threshold is %dC\n", max_temperature);
        }
        else if(!strcmp(argv[i], "-y"))
        {
            if (argc <= i + 1)
                die(false, "The -y option must be followed by a battery capacity percentage\n");
            low_batt_threshold = get_int(argv[++i]);
            errprintf("(cmd line option) Low battery threshold is %u\n", low_batt_threshold);
        }
        // set the three policy choices and test for validity
        else if (!strcmp(argv[i], "-a"))
        {
            if ((argc <= i + 1) || !set_policy(argv[++i], &ac_mode))
                die(false, "The -a option must be followed by a valid policy choice\n");
        }
        else if (!strcmp(argv[i], "-b"))
        {
            if ((argc <= i + 1) || !set_policy(argv[++i], &batt_mode))
                die(false, "The -b option must be followed by a valid policy choice\n");
        }
        else if (!strcmp(argv[i], "-l"))
        {
            if ((argc <= i + 1) || !set_policy(argv[++i], &low_batt_mode))
                die(false, "The -l option must be followed by a valid policy choice\n");
        }
        else if (!strcmp(argv[i], "-A"))
        {
            if ((argc <= i + 1) )
                die(false,
                    "The -A option must be followed by the full path to the acpi ac state file or NULL\n");
            if(!strcmp(argv[++i], "NULL"))
                proc_acpi_ac_state_file = "";
            else
                proc_acpi_ac_state_file = argv[i];
        }
        else if (!strcmp(argv[i], "-I"))
        {
            if ((argc <= i + 1) )
                die(false,
                    "The -I option must be followed by the full path to the acpi battery info file or NULL\n");
            if(!strcmp(argv[++i], "NULL"))
                proc_acpi_batt_info_file = "";
            else
                proc_acpi_batt_info_file = argv[i];
        }
        else if (!strcmp(argv[i], "-S"))
        {
            if ((argc <= i + 1) )
                die(false,
                    "The -S option must be followed by the full path to the acpi battery state file or NULL\n");
            if(!strcmp(argv[++i], "NULL"))
                proc_acpi_batt_state_file = "";
            else
                proc_acpi_batt_state_file = argv[i];
        }
        else if (!strcmp(argv[i], "-T"))
        {
            if ((argc <= i + 1) )
                die(false,
                    "The -T option must be followed by the full path to the acpi temperature file or NULL\n");
            if(!strcmp(argv[++i], "NULL"))
                proc_acpi_temperature_file = "";
            else
                proc_acpi_temperature_file = argv[i];
        }
        else
        {
            fprintf
            (
                stderr,
                "\n%s v%s\n"
                "\n"
                "This program provides a means to vary the CPU's clock speed based on a\n"
                "variety of parameters and inputs.  Available policies are maximum freq,\n"
                "minimum freq, and dynamic frequency control based on the CPU's idle\n"
                "percentage.  Policies may be chosen automatically based on ac online/offline\n"
                "status, remaining battery charge, and CPU temperature (acpi only).  They may\n"
                "also be forced by sending signals to this program."
                "\n\n"
                "Usage: %s [Options]\n"
                "\n"
                "    Options:\n"
                "        -d\n"
                "            Tells the process to daemonize itself (run in background).\n"
                "\n"
                "        -i <interval>\n"
                "            Sets the interval between idle percentage tests and possible\n"
                "            speed changes in 1/10 second increments (default is 20).\n"
                "            Battery status, AC line state, and temperature will be tested\n"
                "            at 5x this interval.\n"
                "\n"
                "        -p <fast up> <threshold>\n"
                "            Sets the CPU idle percentage thresholds.  <fast up> is the idle\n"
                "            percentage below which the CPU will be set to the highest possible\n"
                "            speed.  <threshold> is the idle percentage above which the CPU\n"
                "            speed will be decreased and below which the CPU speed will be\n"
                "            increased (defaults are 10 and 25).\n"
                "\n"
                "        -t  <maxtemp>\n"
                "            Sets the temperature in degrees C, at which CPU will be set to\n"
                "            minimum speed.  Disable temperature monitoring by setting to 0\n"
                "            or a high number.\n"
                "\n"
                "        The following options (a, b, l) can each take one of three policies:\n"
                "        minimum, maximum, dynamic.\n"
                "\n"
                "        -a  <policy>\n"
                "            Sets the policy to use when connected to AC power.  Default is\n"
                "            dynamic.\n"
                "\n"
                "        -b  <policy>\n"
                "            Sets the policy to use when running on battery power.  Default is\n"
                "            minimum.\n"
                "\n"
                "        -l  <policy>\n"
                "            Sets the policy to use when battery capacity is low.  Default is\n"
                "            minimum.\n"
                "\n"
                "        -y  <low batt threshold>\n"
                "            Sets the battery remaining charge percentage at which this\n"
                "            program goes to low battery mode\n"
                "\n"
                "        The program will attempt to find the appropriate ACPI files by\n"
                "        using a (somewhat) intelligent process.  If it fails to find\n"
                "        the correct files, use one or more of the following options.\n"
                "        Using NULL instead of a file path disables the corresponding\n"
                "        function in the program and suppresses the file search.\n"
                "\n"
                "        -A <AC file>\n"
                "            Overrides the default ACPI AC adapter state file.\n"
                "\n"
                "        -I <battery info file>\n"
                "            Overrides the default ACPI battery info file.\n"
                "\n"
                "        -S <battey state file>\n"
                "            Overrides the default ACPI battery state file.\n"
                "\n"
                "        -T <temperature file>\n"
                "            Overrides the default ACPI temperature file.\n"
                "\n"
                "    To tell the program to stay at the highest CPU clock speed to maximize\n"
                "    performance, send the process the SIGUSR1 signal.\n"
                 "\n"
                "    To tell the program to stay at the lowest CPU clock speed to maximize\n"
                "    battery life, send the process the SIGUSR2 signal.\n"
                 "\n"
                "    To tell the program to stay in dynamic scaling mode send the process\n"
                "    the SIGHUP signal.\n"
                "\n"
                "    To tell the program to go back to using its policies send the process\n"
                "    the SIGIO signal.\n"
                "\n"
                "Example:\n"
                "    cpuspeed2 -d -i 10 -t 79 -T /proc/acpi/thermal_zone/THM0/temperature\n"
                 "\n"
                "    This will have the program check (and possibly change) the CPU speed once\n"
                "    every second and go to minimum CPU speed if the temperature goes above 79\n"
                "    using the specified acpi temperature file.\n"
                "\n"
                "Author:\n"
                "    Bill Waddington <william.waddington@beezmo.com>.\n"
                "    Based on Carl Thompson's cpuspeed. (carlthompson.net)\n"
                "\n"
                , NAME, VERSION, NAME
            );
            exit(0);
        }
    }

    // apm or acpi?
    if (!access(PROC_ACPI_DIR, F_OK)) {
        errprintf("Using ACPI\n");
        using_acpi = true;
    } else
        errprintf("Using APM\n");

    // if acpi mode and user hasn't provided file options,
    // build the paths
    if(using_acpi) {
        if(!proc_acpi_ac_state_file) {
             find_path(PROC_ACPI_AC_STATE_BASE, PROC_ACPI_AC_STATE_FILE,
                 acpi_ac_state_path, MAX_PATH_LEN);
             proc_acpi_ac_state_file = acpi_ac_state_path;
        }
        if(!proc_acpi_batt_state_file) {
             find_path(PROC_ACPI_BATT_STATE_BASE, PROC_ACPI_BATT_STATE_FILE,
                 acpi_batt_state_path, MAX_PATH_LEN);
             proc_acpi_batt_state_file = acpi_batt_state_path;
        }
        if(!proc_acpi_batt_info_file) {
             find_path(PROC_ACPI_BATT_INFO_BASE, PROC_ACPI_BATT_INFO_FILE,
                 acpi_batt_info_path, MAX_PATH_LEN);
             proc_acpi_batt_info_file = acpi_batt_info_path;
        }
        if(!proc_acpi_temperature_file) {
             find_path(PROC_ACPI_TEMPERATURE_BASE, PROC_ACPI_TEMPERATURE_FILE,
                 acpi_temperature_path, MAX_PATH_LEN);
             proc_acpi_temperature_file = acpi_temperature_path;
        }
        errprintf("ACPI AC state file = %s\n", proc_acpi_ac_state_file);
        errprintf("ACPI battery state file = %s\n", proc_acpi_batt_state_file);
        errprintf("ACPI battery info file = %s\n", proc_acpi_batt_info_file);
        errprintf("ACPI temperature file = %s\n", proc_acpi_temperature_file);
    }

    // use the userspace governor
    write_line(SYSFS_GOVERNOR_FILE, "%s\n", SYSFS_USERSPACE);

    if (access(SYSFS_CURRENT_SPEED_FILE, W_OK) < 0)
        die(true, "Cannot write to speed control file: %s", SYSFS_CURRENT_SPEED_FILE);

    // run in background if requested
    if (daemonize)
        daemon(0, 0);

    // set up signal handling
    struct sigaction signal_action;

    // block all of our signals during each signal handler to avoid possible
    // race condition (not that it would really matter)
    sigemptyset(&signal_action.sa_mask);
    sigaddset(&signal_action.sa_mask, SIGALRM);
    sigaddset(&signal_action.sa_mask, SIGUSR1);
    sigaddset(&signal_action.sa_mask, SIGUSR2);
    sigaddset(&signal_action.sa_mask, SIGHUP);
    sigaddset(&signal_action.sa_mask, SIGIO);
    signal_action.sa_flags = 0;

    // set the SIGALRM handler to our function
    signal_action.sa_handler = alarm_handler;
    sigaction(SIGALRM, &signal_action, 0);

    // set the SIGUSR1 handler to our function
    signal_action.sa_handler = usr1_handler;
    sigaction(SIGUSR1, &signal_action, 0);

    // set the SIGUSR2 handler to our function
    signal_action.sa_handler = usr2_handler;
    sigaction(SIGUSR2, &signal_action, 0);

    // set the HUP handler to our function
    signal_action.sa_handler = hup_handler;
    sigaction(SIGHUP, &signal_action, 0);

    // set the IO handler to our function
    signal_action.sa_handler = io_handler;
    sigaction(SIGIO, &signal_action, 0);

    // reset the speed steps
    get_supported_speeds();
    // reset the time counters
    reset_times();

    // we dynamically scale speed to start
    mode = DYNAMIC;

    // and use our policies unless a forcing signal rx'd
    sig_mode = SIG_FREE;

    // and assume AC for now
    power_therm_status = AC;

    // set working interval (may have changed from default)
    working_interval = interval;

    // main loop
    while (1)
    {
        raise(SIGALRM);
        usleep(working_interval * 100000L);
    }

    // never reached
    return 0;
}
